/*
 *  ChannelWindow.cpp - The window type for channels
 *  Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Python.h>

#include <stdlib.h>
#include <Integer.h>

#include "PyIRC.h"
#include "User.h"
#include "ChannelWindow.h"
#include "Util.h"
#include "PopMenu.h"

extern PyIRC *mainWindow;

QPopupMenu * ChannelWindow::rightMenu = 0;

ChannelWindow::ChannelWindow(String channelname)
  : LoggingWindow(channelname), topic(""), key(""), limit(0),
    mode(0), status(JOINING)
{
  messageArea->setGeometry(0, 30, 400, 310);
  nickListBox = new QListBox(this, "NickList");
  nickListBox->setGeometry(400, 30, 100, 310);

  nickListBox->installEventFilter(this);
}

ChannelWindow::~ChannelWindow()
{
  delete nickListBox;
}

void
ChannelWindow::handleInput()
{
  String message = messageEdit->text();

  if (message == "")
    return;

  messageEdit->setText("");

  if (message[0]=='/')
    Parser::parseCommand(getName(), message);
  else
    if (mainWindow->registered) {
      Util::send(mainWindow->getSock(), String("PRIVMSG ") + getName() +
        " :" + message);
      Util::do_on(3, "SEND_PUBLIC", getName().chars(),
        message.chars());
    }
}

void
ChannelWindow::resizeEvent(QResizeEvent *re)
{
  messageArea->resize(re->size().width() - 100,
                      re->size().height() - 52);
  nickListBox->setGeometry(re->size().width() - 100,
                        30, 100,
                        re->size().height() - 52);
  messageEdit->setGeometry(0, re->size().height() - 22,
                           re->size().width(), 22);
  messageArea->setPositionToEnd();
}

bool
ChannelWindow::eventFilter(QObject *obj, QEvent *ev)
{
  if (obj == nickListBox) {
    if (ev->type() == Event_MouseButtonPress) {

      if (rightMenu == 0) {
        rightMenu = new PopMenu("Nick-related Ops");
        rightMenu->insertItem("Give/Remove &Ops");
        rightMenu->insertItem("&Kick");
      }

      QMouseEvent *e = (QMouseEvent *)ev;
      if (e->button() == RightButton && rightMenu != 0)
        rightMenu->popup(QCursor::pos());
    }
  }
  return FALSE;
}

void
ChannelWindow::addNick(String n, int pos)
{
  User *u;
  String realnick;

  nickListBox->insertItem(n, pos);

  if (n[0] == '@' || n[0] == '+')
    realnick = n.after(0);
  else
    realnick = n;

  u = new User(realnick);
  if (n[0] == '@')
    u->setChanop(TRUE);
  if (n[0] == '+')
    u->setVoice(TRUE);

  nickList.append(u);
}

int
ChannelWindow::findNick(String n)
{
  int i;
  User *u;
  QListIterator<User> it(nickList);

  for (i = 0; u = it.current(); ++it, ++i) {
    if (u->nickname == n) {
      return i;
    }
  }
  return -1;
}

bool
ChannelWindow::hasNick(String nick)
{
  return (findNick(nick) != -1);
}

void
ChannelWindow::delNick(String n)
{
  int i = findNick(n);
  nickList.remove(i);
  nickListBox->removeItem(i);
}

void
ChannelWindow::clearNicks()
{
  while (nickList.count() != 0) {
    nickList.remove((unsigned int)0);
    nickListBox->removeItem(0);
  }
}

void
ChannelWindow::changeNick(String oldnick, String newnick)
{
  int i;
  User *u = nickList.at(i = findNick(oldnick));
  u->nickname = newnick;
  nickListBox->changeItem(u->toString().chars(), i);
}

void
ChannelWindow::setOper(String nick, bool oper)
{
  int i = findNick(nick);
  if (i == -1)
    return;
  nickList.at(i)->setChanop(oper);
  nickListBox->changeItem(nickList.at(i)->toString().chars(), i);
}

bool
ChannelWindow::isOper(String nick)
{
  int i = findNick(nick);
  if (i == -1)
    return FALSE;
  return nickList.at(i)->chanop;
}

void
ChannelWindow::setVoice(String nick, bool oper)
{
  int i = findNick(nick);
  nickList.at(i)->setVoice(oper);
  nickListBox->changeItem(nickList.at(i)->toString().chars(), i);
}

void
ChannelWindow::setTopic(String t)
{
  topic = Util::removeNonAlpha(t);
  newTitle();
}

void
ChannelWindow::newTitle(void)
{
  String t = "PyIRC - " + getName();
  String u = "";

  if (mode != 0) {
    t = t + " (+";
    if (mode & INVITE_ONLY) t += "i";
    if (key != "") { u = u + " " + dec(limit); t += "k"; }
    if (limit != 0) { u += " " + key; t += "l"; }
    if (mode & MODERATED) t += "m";
    if (mode & EXTMSG_RESTRICTED) t += "n";
    if (mode & PRIVATE) t += "p";
    if (mode & SECRET) t += "s";
    if (mode & TOPIC_RESTRICTED) t += "t";
    if (u != "") t+= u;
    t += ")";
  }

  if (topic != "")
    t = t + " [" + topic + "]";

   setCaption(t);
}

void
ChannelWindow::parseMode(String m, String p)
{
  char sign = '-';

  for (unsigned int i = 0; i < m.length(); i++)
    switch(m[i]) {
      case '+':
      case '-':
        sign = m[i];
        break;
      case 'p': sign =='+' ? mode |= PRIVATE : mode &= ~PRIVATE; break;
      case 's': sign =='+' ? mode |= SECRET : mode &= ~SECRET; break;
      case 'i': sign =='+' ? mode |= INVITE_ONLY :
                             mode &= ~INVITE_ONLY; break;
      case 't': sign =='+' ? mode |= TOPIC_RESTRICTED :
                             mode &= ~TOPIC_RESTRICTED; break;
      case 'n': sign =='+' ? mode |= EXTMSG_RESTRICTED :
                             mode &= ~EXTMSG_RESTRICTED; break;
      case 'm': sign =='+' ? mode |= MODERATED :
                             mode &= ~MODERATED; break;
      case 'l':
        if (sign == '+') {
          limit = atoi(p.before(" ").chars());
          p = p.after(" ");
        } else
          limit = 0;
        break;
      case 'k':
        if (sign == '+')
          key = p.before(" ");
        else
          key = "";
        p = p.after(" ");
        break;
      case 'o':
        setOper(p.before(" "), (bool) (sign=='+'));
        p = p.after(" ");
        break;
      case 'v':
        setVoice(p.before(" "), (bool) (sign=='+'));
        p = p.after(" ");
        break;
      case  'b':
        p = p.after(" ");
    }
  newTitle();
}
