/*
 *  ConnectDialog.cpp - The connection dialog box
 *  Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <fstream.h>
#include <Integer.h>

#include "ConnectDialog.h"
#include "Util.h"
#include "PyIRC.h"

extern PyIRC * mainWindow;

ConnectDialog::ConnectDialog(QWidget *parent, const char *name)
  : QTabDialog(parent, name, TRUE)
{
  QLabel *anyLabel;

  setCaption("PyIRC - Server and nickname selection");

  configFile = "";

  // Set up the server page
  QWidget *w = new QWidget(this, "server page");

  anyLabel = new QLabel(w, "label server name");
  anyLabel->setGeometry(240, 10, 100, 20);
  anyLabel->setText("Server address:");
  servernameEdit = new QLineEdit(w, "server name editor");
  servernameEdit->setGeometry(240, 35, 138, 25);
  connect(servernameEdit, SIGNAL(textChanged(const char*)),
          this, SLOT(enableAddServer()));

  anyLabel = new QLabel(w, "label server port");
  anyLabel->setGeometry(240, 75, 100, 20);
  anyLabel->setText("Server port:");
  serverportEdit = new QLineEdit(w, "server port editor");
  serverportEdit->setGeometry(240, 100, 138, 25);
  connect(serverportEdit, SIGNAL(textChanged(const char*)),
          this, SLOT(enableAddServer()));

  addButton = new QPushButton(w, "add button");
  addButton->setGeometry(260, 170, 118, 25);
  addButton->setText("Add server");
  addButton->setEnabled(FALSE);
  connect(addButton, SIGNAL(clicked()), this, SLOT(addServer()));

  removeButton = new QPushButton(w, "remove button");
  removeButton->setGeometry(260, 200, 118, 25);
  removeButton->setText("Remove server");
  connect(removeButton, SIGNAL(clicked()), this, SLOT(removeServer()));

  serverListBox = new QListBox(w, "server listbox");
  serverListBox->setGeometry(10, 10, 220, 215);
  serverListBox->setFrameStyle(QFrame::WinPanel | QFrame::Sunken);
  connect(serverListBox, SIGNAL(highlighted(int)), this,
          SLOT(selectedServer(int)));

  addTab(w, "Server");

  // Set up the nickname page
  w = new QWidget(this, "nick page");

  anyLabel = new QLabel(w, "label nick");
  anyLabel->setGeometry(10, 10, 105, 25);
  anyLabel->setAlignment(AlignRight | AlignVCenter);
  anyLabel->setText("Nickname:");
  nicknameEdit = new QLineEdit(w, "nick line editor");
  nicknameEdit->setGeometry(120, 10, 260, 25);
  connect(nicknameEdit, SIGNAL(textChanged(const char*)),
          this, SLOT(modifyNickname()));

  anyLabel = new QLabel(w, "label username");
  anyLabel->setGeometry(10, 45, 105, 25);
  anyLabel->setAlignment(AlignRight | AlignVCenter);
  anyLabel->setText("Username:");
  usernameEdit = new QLineEdit(w, "username line editor");
  usernameEdit->setGeometry(120, 45, 260, 25);
  connect(usernameEdit, SIGNAL(textChanged(const char*)),
          this, SLOT(modifyNickname()));

  anyLabel = new QLabel(w, "label ircname");
  anyLabel->setGeometry(10, 80, 105, 25);
  anyLabel->setAlignment(AlignRight | AlignVCenter);
  anyLabel->setText("Ircname/Realname:");
  ircnameEdit = new QLineEdit(w, "ircname line editor");
  ircnameEdit->setGeometry(120, 80, 260, 25);
  connect(ircnameEdit, SIGNAL(textChanged(const char*)),
          this, SLOT(modifyNickname()));

  addTab(w, "Nickname");

  // Set up the settings page
  w = new QWidget(this, "settings page");

  saveButton = new QPushButton(w, "save button");
  saveButton->setGeometry(140, 80, 100, 25);
  saveButton->setText("Save");
  saveButton->setEnabled(FALSE);
  connect(saveButton, SIGNAL(clicked()),
          this, SLOT(saveConfigFile()));

  loadButton = new QPushButton(w, "load button");
  loadButton->setGeometry(140, 110, 100, 25);
  loadButton->setText("Load");
  connect(loadButton, SIGNAL(clicked()),
          this, SLOT(reloadConfigFile()));

  addTab(w, "Settings");
  
  setCancelButton();

  serverList.setAutoDelete(TRUE);
  connect(this, SIGNAL(applyButtonPressed()), SLOT(apply()));

  setFixedSize(400, 300);
}

ConnectDialog::~ConnectDialog()
{
  delete serverListBox;
  delete nicknameEdit;
  delete usernameEdit;
  delete ircnameEdit;
  delete servernameEdit;
  delete serverportEdit;
  delete addButton;
  delete removeButton;
  delete saveButton;
  delete loadButton;
}

void
ConnectDialog::apply()
{
  if (nicknameEdit->text()!="" &&
      usernameEdit->text()!="" &&
      serverListBox->currentItem() != -1) {
    mainWindow->setNickname(nicknameEdit->text());
    mainWindow->setUsername(usernameEdit->text());
    mainWindow->setIrcname(ircnameEdit->text());
    mainWindow->setServer(serverList.at(serverListBox->currentItem()));
    mainWindow->ircConnect();
  }
}

struct keyword {
  char *key;
  void (ConnectDialog::*function)(String);
} keywords[] =
{
  { "NICKNAME",	ConnectDialog::config_nickname	},
  { "NICK",	ConnectDialog::config_nickname	},
  { "USERNAME",	ConnectDialog::config_username	},
  { "USER",	ConnectDialog::config_username	},
  { "IRCNAME",	ConnectDialog::config_ircname	},
  { "REALNAME",	ConnectDialog::config_ircname	},
  { "SERVER",	ConnectDialog::config_server	},
  { 0,		0					}
};

void
ConnectDialog::readConfigFile(String filename)
{
  String s, t, u;
  int i;

  configFile = filename;

  fstream file((const char *)configFile, ios::in | ios::out);
  while (!file.eof()) {

    s="";
    readline(file, s);

    s = Util::removeLeadingBlanks(s);

    if (s.length()==0 || s[0]=='#')
      continue;

    t = s.before("=");

    if (t.length()==0) {
      cerr << "Syntax error. Expecting '='" << endl;
      cerr << "Aborting." << endl;
      exit(1);
    }

    t = upcase(Util::removeTrailingBlanks(t));
    u = Util::removeTrailingBlanks(Util::removeLeadingBlanks(s.after("=")));

    for (i = 0; keywords[i].key != 0; i++)
      if (t==keywords[i].key) {
        (keywords[i].function)(u);
        break;
      }

    if (keywords[i].key==0) {
      cerr << "Syntax error. No keyword for '" << t << "'.\n";
      cerr << "Aborting.\n";
      exit(1);
    }
  }
  if (serverListBox->count() != 0)
    serverListBox->setCurrentItem(0);

  file.close();
}

void
ConnectDialog::saveConfigFile()
{
  unsigned int i = 0;

  ofstream file((const char *)configFile);
  file << "# PyIRC configuration file\n\n# The nickname we will try to use\nNICKNAME = ";
  file << nicknameEdit->text() << "\n\n# Our username (remember, nick!user@host)\nUSERNAME = ";
  file << usernameEdit->text() << "\n\n# This is our real name\nIRCNAME = ";
  file << ircnameEdit->text() << "\n\n#The server list.\n";

  while (i < serverList.count())
    file << "SERVER = " << serverList.at(i)->name << " "
         << serverList.at(i++)->port << "\n";

  file.close();
  saveButton->setEnabled(FALSE);
}

void
ConnectDialog::reloadConfigFile(void)
{
  serverListBox->clear();
  while (serverList.count() != 0)
    serverList.remove((unsigned int)0);
  readConfigFile(configFile);
}

void
ConnectDialog::config_nickname(String param)
{
  nicknameEdit->setText(param);
}

void
ConnectDialog::config_username(String param)
{
  usernameEdit->setText(param);
}

void
ConnectDialog::config_ircname(String param)
{
  ircnameEdit->setText(param);
}

void
ConnectDialog::config_server(String param)
{
  String words[2], str;

  if (split(param, words, 2, RXwhite)!=2 ||
      atoi((const char *)words[1])==0) {
    cerr << "The server definition must respect the form:\n";
    cerr << "SERVER = <server name> <port number>\n";
    exit(1);
  }
  serverList.append(new Server(words[0], atoi((const char *)words[1])));
  str = words[0] + ", " + words[1];
  serverListBox->insertItem((const char *)str);
}

void
ConnectDialog::modifyNickname()
{
  saveButton->setEnabled(TRUE);
}

void
ConnectDialog::addServer()
{
  Server *serv = new Server(servernameEdit->text(),
                            atoi(serverportEdit->text()));
  serverList.append(serv);
  serverListBox->insertItem(serv->name + ", " + dec(serv->port));
  addButton->setEnabled(FALSE);
  saveButton->setEnabled(TRUE);
}

void
ConnectDialog::removeServer()
{
  int i = serverListBox->currentItem();
  serverListBox->removeItem(i);
  serverList.remove(i);
  if (serverListBox->count() == 0)
    removeButton->setEnabled(FALSE);
  else
    selectedServer(serverListBox->currentItem());
  saveButton->setEnabled(TRUE);
}

void
ConnectDialog::selectedServer(int index)
{
  servernameEdit->setText((const char *)((serverList.at(index))->name));
  serverportEdit->setText(dec((serverList.at(index))->port));
  addButton->setEnabled(FALSE);
}

void
ConnectDialog::enableAddServer()
{
  addButton->setEnabled(TRUE);
}
