/*
 *  LineEdit.cpp - A customizable line editor, based on Qt's line editor.
 *  Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr)
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <qfontmet.h>
#include <qpixmap.h>
#include <qkeycode.h>
#include "LineEdit.h"

LineEdit::LineEdit(QWidget *parent, const char *name)
  : QWidget(parent, name), buf(""), cursorPos(0), maxLen(32767),
    offset(0)
{
  initMetaObject();
  setFocusPolicy(StrongFocus);
  setCursor(ibeamCursor);
}

LineEdit::~LineEdit()
{
}

const char *
LineEdit::text() const
{
  return buf.chars();
}

void
LineEdit::setText(const char *text)
{
  if (buf == text)
    return;
  buf = text ? text : "";
  if (buf.length() > maxLen)
    buf = buf.before((int)maxLen);
  cursorPos = 0;
  repaint(!hasFocus());
  emit textChanged(buf.chars());
}

int
LineEdit::maxLength() const
{
  return maxLen;
}

void
LineEdit::setMaxLength(int len)
{
  maxLen = len;
  cursorPos = 0;
  if (buf.length() > maxLen)
    buf = buf.before((int)maxLen);
  repaint(!hasFocus());
}

void
LineEdit::cursorLeft(int steps)
{
  if (steps < 0) {
    cursorRight(-steps);
    return;
  }

  if (cursorPos != 0) {
    cursorPos -= steps;
    if (cursorPos < 0)
      cursorPos = 0;
    if (cursorPos < offset)
      offset = cursorPos;
    repaint(!hasFocus());
  }
}

void
LineEdit::cursorRight(int steps)
{
  if (steps < 0) {
    cursorLeft(-steps);
    return;
  }

  int len = buf.length();
  if (cursorPos != len) {
    cursorPos += steps;
    if (cursorPos > len)
      cursorPos = len;
    QFontMetrics fm = fontMetrics();
    int diff = width() - 4 - fm.width(buf.after(offset).chars(),
                 cursorPos - offset);
    while (diff < 0 && offset < cursorPos - 1) {
      diff += fm.width(buf[offset]);
      offset++;
    }

    repaint(!hasFocus());
  }
}

void
LineEdit::backspace(void)
{
  if (cursorPos > 0) {
    cursorLeft();
    del();
  }
}

void
LineEdit::del(void)
{
  if (cursorPos != buf.length()) {
    buf = buf.before(cursorPos) + buf.after(cursorPos + 1);
    repaint(!hasFocus());
    emit textChanged(buf.chars());
  }
}

void
LineEdit::begin(void)
{
  cursorPos = offset = 0;
  repaint(!hasFocus());
}

void
LineEdit::end(void)
{
  cursorRight(buf.length() - cursorPos + 1);
}

void
LineEdit::keyPressEvent(QKeyEvent *e)
{
  if (e->key() == Key_Enter || e->key() == Key_Return) {
    emit returnPressed();
    return;
  }
  if (e->key() != Key_Delete && e->ascii() != 0) {
    if (buf.length() < maxLen) {
      buf = buf.before(cursorPos) + e->ascii() + buf.after(cursorPos);
      cursorRight();
    }
  }
}

void
LineEdit::focusInEvent(QFocusEvent *)
{
}

void
LineEdit::focusOutEvent(QFocusEvent *)
{
}

void
LineEdit::paintEvent(QPaintEvent *)
{
  QPainter p;
  p.begin(this);
  QColorGroup cg = colorGroup();
  QColor bg = isEnabled() ? cg.base() : cg.background();
  p->fillRect(2, 2, width() - 4, height() - 4, bg);
  String t = buf.after(cursorPos);
  p.end();
}

void
LineEdit::resizeEvent(QResizeEvent *)
{
}
