#include <Integer.h>

#include <qpainter.h>
#include <qlabel.h>
#include <qtooltip.h>
#include <qpushbt.h>

#include "LinksWindow.h"

#define MIN(a,b) ((a)<(b) ? (a) : (b))
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define ABS(a)   ((a)<0 ? (-(a)): (a))

class QLine : public QWidget
{
public:
  QLine(QWidget *parent = 0, const char *name = 0)
    : QWidget(parent, name) { xmin = xmax = ymin = ymax = 0; }
  setLine(int x1, int y1, int x2, int y2) {
    setGeometry(MIN(x1, x2), MIN(y1, y2),
                ABS(x2-x1), ABS(y2-y1));
    xmin = MAX(0, x2-x1); ymin = MAX(0, y2-y1);
    xmax = MAX(0, x1-x2); ymax = MAX(0, y1-y2);
  }

protected:
  int xmin, xmax, ymin, ymax;

  void paintEvent(QPaintEvent*)
    {  QPainter p(this); p.drawLine(xmin, ymin, xmax, ymax);  }
};

LinksWindow::LinksWindow(QWidget *parent, const char *name)
  : QDialog(parent, name, FALSE)
{
  setFixedSize(400, 335);

  nodeList.setAutoDelete(TRUE);
  widgetList.setAutoDelete(TRUE);

  fm = new QFontMetrics(font());

  status = LinksWindow::RECEIVING;

  QPushButton *closeButton = new QPushButton(this, "close button");
  closeButton->setText("Close");
  closeButton->setGeometry(150, 305, 100, 25);
  closeButton->setDefault(TRUE);
  connect(closeButton, SIGNAL(clicked()), this, SLOT(accept()));
}

LinksWindow::~LinksWindow()
{
  delete fm;
}

void
LinksWindow::rescale(void)
{
  // We recenter
  double x = 0, y = 0;
  QListIterator<Node> it(nodeList);
  Node *n;
  for ( ; n = it.current(); ++it) {
    x += n->x; y += n->y;
  }
  x = x / nodeList.count();
  y = y / nodeList.count();
  it.toFirst();
  for ( ; n = it.current(); ++it) {
    n->x -= x; n->y -= y;
  }

  // We rotate
  // To find the angle, we compute the inertia matrice, then
  // the eigenvector associated to the biggest eigenvalue.
  double xx = 0.0, yy = 0.0, xy = 0.0;
  it.toFirst();
  for ( ; n = it.current(); ++it) {
    xx = xx + n->y * n->y;
    yy = yy + n->x * n->x;
    xy = xy - n->x * n->y;
  }

  double lambda = (xx+yy+sqrt((xx + yy)*(xx + yy)-4*(xx*yy-xy*xy)))/2.0;
  // -theta is the angle between the longest eigenvector and the
  // horizontal
  double theta = -asin(1/sqrt(1+((lambda-xx)/xy)*((lambda-xx)/xy)));

  it.toFirst();
  for ( ; n = it.current(); ++it) {
    x = cos(theta) * n->x + sin(theta) * n->y;
    n->y = - sin(theta) * n->x + cos(theta) * n->y;
    n->x = x;
  }

  // We rescale;
  it.toFirst();
  x = 0;
  for ( ; n = it.current(); ++it)
    if (x < (n->x * n->x + n->y * n->y))
      x = n->x * n->x + n->y * n->y;
  x = 145.0 / sqrt(x);
  it.toFirst();
  for ( ; n = it.current(); ++it) {
    n->x *= x; n->y *= x;
  }
}

void
LinksWindow::rotateNodes(Node *n, double vx, double vy)
{
  n->fixed = TRUE;

  if (n->connectedTo.count() == 1) {
    if (n->connectedTo.at(0)->fixed == FALSE)
      rotateNodes(n->connectedTo.at(0), vx, vy);
  } else {
    double alpha = 2 * M_PI / n->connectedTo.count();
    int j = 0;

    QListIterator<Node> it(n->connectedTo);
    Node *m;
    for (; m = it.current(); ++it)
      if (m->fixed == FALSE) {
        j++;
        m->x = n->x - cos(alpha * j) * vx + sin(alpha * j) * vy;
        m->y = n->y - sin(alpha * j) * vx - cos(alpha * j) * vy;
        rotateNodes(m, m->x - n->x, m->y - n->y);
      }
  }
}

void
LinksWindow::addLink(String from, String to)
{
  if (status == LinksWindow::FINISHED) {
    nodeList.clear();
    widgetList.clear();
    erase();
    status = LinksWindow::RECEIVING;
  }

  Node *newNode = new Node();
  nodeList.append(newNode);
  newNode->connectedTo.clear();
  newNode->x = newNode->y = 0;
  newNode->name = from;
  newNode->fixed = FALSE;

  QListIterator<Node> it(nodeList);
  Node *m;

  if (to != from) {
    for ( ; m = it.current(); ++it)
      if (m->name == to) {
        m->connectedTo.append(newNode);
        newNode->connectedTo.append(m);
        newNode->x = m->x - 1;
        newNode->y = m->y;
        break;
      }
    rotateNodes(newNode, 1, 0);
    it.toFirst();
  }

  for ( ; m = it.current(); ++it)
    m->fixed = FALSE;
}

void
LinksWindow::create(Node *n)
{
  int w;

  n->fixed = TRUE;

  QListIterator<Node> it(n->connectedTo);
  Node *m;
  for (; m = it.current(); ++it)
    if (m->fixed == FALSE) {
      QLine *line = new QLine(this, "a line");
      line->setLine(200 + (int) (n->x),
                    150 + (int) (n->y),
                    200 + (int) (m->x),
                    150 + (int) (m->y));
      widgetList.append(line);
      create(m);
    }

  QLabel *anyLabel = new QLabel(this, "a server");
  anyLabel->setFrameStyle(QFrame::Box | QFrame::Sunken);
  anyLabel->setAlignment(AlignCenter);
  anyLabel->setText(dec(nodeList.find(n) + 1));
  w = fm->width(anyLabel->text());
  anyLabel->setGeometry(200 + (int)(n->x) - w, 150 + (int)(n->y) -
                        fm->height() / 2 - 1 , w * 2, fm->height() + 2);
  widgetList.append(anyLabel);
  QToolTip::add(anyLabel, n->name.chars());
}

void
LinksWindow::doneLinks(void)
{
  status = LinksWindow::FINISHED;
  rescale();
  show();
  repaint();
}

void
LinksWindow::show()
{
  widgetList.clear();
  erase();
  create(nodeList.at(0));
  QDialog::show();
}
