/*
 *  Parser.cpp - The server-side parser
 *  Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Python.h>

#include <qlist.h>
#include <time.h>
#include <stdlib.h>

#include "Parser.h"
#include "PyIRC.h"
#include "Util.h"
#include "ChannelWindow.h"
#include "Defines.h"

extern PyIRC * mainWindow;
extern PyObject *globalDict;

PyObject *windowName = 0;
PyObject *nickName = 0;

Parser::Parser(int sock)
{
  nick = "";
  socket = sock;
}

Parser::~Parser()
{
}

struct parseMessage {
  char *name;
  void (Parser::*function)(String, String);
} parseMessages[] = {
  { "PRIVMSG",	Parser::parsePrivmsg	},
  { "ERROR",	Parser::parseError	},
  { "PING",	Parser::parsePing	},
  { "NOTICE",	Parser::parseNotice	},
  { "JOIN",	Parser::parseJoin	},
  { "TOPIC",	Parser::parseTopic	},
  { "MODE",	Parser::parseMode	},
  { "PART",	Parser::parsePart	},
  { "KICK",	Parser::parseKick	},
  { "KILL",	Parser::parseKill	},
  { "QUIT",	Parser::parseQuit	},
  { "NICK",	Parser::parseNick	},
  { "001",	Parser::parse001	},
  { "324",	Parser::parse324	},
  { "332",	Parser::parse332	},
  { "353",	Parser::parse353	},
  { "364",	Parser::parse364	},
  { "365",	Parser::parse365	},
  { "366",	Parser::parse366	},
  { "432",	Parser::parse432	},
  { "433",	Parser::parse433	},
  { 0,		0			}
};

void
Parser::parseInput(String str)
{
  String from = "", command, parameters, p;
  int i;

  if (Util::do_on(2, "RAW_IRC", str.chars()))
    return;

  if (str[0] == ':') {
    str = str.after(":");
    from = str.before(" ");
    str = str.after(" ");
  }

  command = upcase(str.before(" "));
  parameters = str.after(" ");

  for (i = 0; parseMessages[i].name != 0; i++)
    if (command == parseMessages[i].name) {
      (parseMessages[i].function)(from, parameters);
      return;
    }

  // Numerics
  if (atoi(command) > 0 && atoi(command) < 999) {
    p = parameters.after(" ");
    if (Util::do_on(2, command.chars(), p.chars()))
      return;
  }

  cerr << "Warning, unknown state :" << command << endl;
  cerr << command << " " << parameters << endl;
}

void
Parser::parseError(String from, String parameters)
{
  mainWindow->ircDisconnect();
}

void
Parser::parseJoin(String from, String parameters)
{
  String n, channel;

  channel = downcase(parameters.after(":"));
  n = from.before("!");
  Util::do_on(3, "JOIN", from.chars(), channel.chars());
  if (n == nick)
     Util::send(socket, String("MODE ") + channel);
  else {
    ChannelWindow *w = (ChannelWindow *)mainWindow->findWindow(channel);
    w->addNick(n, 0);
  }
}

void
Parser::parsePing(String from, String parameters)
{
  Util::send(socket,
             "PONG %s",(const char *)parameters);
}

void
Parser::parseCTCP(String fromnick, String to, String command,
                  String parameters)
{
  Util::do_on(5, "CTCP", fromnick.chars(), to.chars(),
              command.chars(), parameters.chars());
}

void
Parser::parsePrivmsg(String from, String parameters)
{
  String to, message, n;

  to = downcase(parameters.before(" "));
  message = parameters.after(":");
  n = from.before("!");

  if (message[0] == '\001' && message[message.length()-1] == '\001') {
    message = message.after(0);
    message = message.before((int)(message.length() - 1));
    if (message.before(" ") == "")
      parseCTCP(n, to, message, "");
    else
      parseCTCP(n, to, message.before(" "), message.after(" "));
  } else if (to.index('*') != -1)
    Util::do_on(4, "MSG_GROUP", n.chars(), to.chars(),
                message.chars());
  else if (to[0] == '#' || to[0] == '&') {
    ChannelWindow *w = (ChannelWindow *)mainWindow->findWindow(to);
    if (w->hasNick(n))
      Util::do_on(4, "PUBLIC", n.chars(), to.chars(),
                  message.chars());
    else
      Util::do_on(4, "PUBLIC_MSG", n.chars(), to.chars(),
                  message.chars());
  } else
    Util::do_on(3, "MSG", n.chars(), message.chars());
}

void
Parser::parseNotice(String from, String parameters)
{
  String to, message, n;

  to = downcase(parameters.before(" "));
  message = parameters.after(":");

  if (from.index('!') == -1)
    Util::do_on(3, "SERVER_NOTICE", from.chars(),
                message.chars());
  else {
    n = from.before("!");
    if (to[0] == '#' || to[0] == '&')
      Util::do_on(4, "PUBLIC_NOTICE", n.chars(), to.chars(),
                  message.chars());
    else
      Util::do_on(3, "NOTICE", n.chars(), message.chars());
  }
}

void
Parser::parse001(String from, String parameters)
{
  mainWindow->registered = TRUE;
  nick = parameters.before(" ");
  Py_XDECREF(nickName);
  PyObject *o = PyImport_AddModule("raw");
  PyObject *p = PyModule_GetDict(o);
  nickName = PyString_FromString((char *)nick.chars());
  PyDict_SetItemString(p, "nickname", nickName);

  mainWindow->loadInitFile();

  Util::do_on(2, "001", parameters.chars());
}

void
Parser::parseTopic(String from, String parameters)
{
  String topic, channel;
  ChannelWindow *w;

  // parameters = "channel :topic"
  channel = downcase(parameters.before(" "));
  topic = parameters.after(":");
  from = from.before("!");

  w = (ChannelWindow *)mainWindow->findWindow(channel);
  w->setTopic(topic);

  Util::do_on(4, "TOPIC", from.chars(), channel.chars(),
              topic.chars());
}

void
Parser::parseMode(String from, String parameters)
{
  String to;
  ChannelWindow *w;

  to = downcase(parameters.before(" "));
  if (to[0] == '#' || to[0]=='&') {
    parameters = parameters.after(" ");
    w = (ChannelWindow *)mainWindow->findWindow(to);
    w->parseMode(parameters.before(" "), parameters.after(" "));
  } else {
    parameters = parameters.after(":");
    mainWindow->parseMode(parameters);
  }

  Util::do_on(4, "MODE", from.chars(), to.chars(),
              parameters.chars());
}

void
Parser::parseQuit(String from, String parameters)
{
  Window *w;
  parameters = parameters.after(":");
  if (from.index("!") != -1)
    from = from.before("!");
  Util::do_on(3, "SIGNOFF", from.chars(), parameters.chars());
  if (from != nick && from.index(".") != -1) {
      QListIterator<Window> it(mainWindow->windowList);
      for (; w = it.current(); ++it)
        if (Util::isChannelName(w->getName())) {
          ChannelWindow *c = (ChannelWindow *)w;
          if (c->hasNick(from))
            c->delNick(from);
        }
  }
}

void
Parser::parsePart(String from, String parameters)
{
  String channel, n, bonus = "";

  n = from.before("!");
  channel = downcase(parameters.before(" :"));
  if (channel == "")
    channel = parameters;
  else
    bonus = parameters.after(":");

  ChannelWindow *w = (ChannelWindow *)mainWindow->findWindow(channel);
#ifdef DBG
  cout << "parsePart : w = " << w << "\n";
#endif
  if (w != 0) {
    if (n == nick)
      mainWindow->windowList.remove(w);
    else
      w->delNick(n);
  }

  Util::do_on(4, "LEAVE", n.chars(), channel.chars(),
              bonus.chars());
}

void
Parser::parseKick(String from, String parameters)
{
  String channel, n, d;

  n = from.before("!");
  channel = downcase(parameters.before(" "));
  parameters = parameters.after(" ");
  d = parameters.before(" ");
  parameters = parameters.after(":");

  ChannelWindow *w = (ChannelWindow *)mainWindow->findWindow(channel);
  if (d == nick) {
    w->clearNicks();
    w->setStatus(ChannelWindow::KICKED);
  } else
    w->delNick(d);

  Util::do_on(5, "KICK", n.chars(), d.chars(),
              channel.chars(), parameters.chars());
}

void
Parser::parseNick(String from, String parameters)
{
  String n = from.before("!");
  parameters = parameters.after(":");

  QListIterator<Window> it(mainWindow->windowList);
  Window *w;
  for (; w = it.current(); ++it)
    if (Util::isChannelName(w->getName())) {
      ChannelWindow *c = (ChannelWindow *)w;
      if (c->hasNick(n))
        c->changeNick(n,parameters);
    }

  Util::do_on(3, "NICKNAME", n.chars(), parameters.chars());

  if (n == nick) {
    nick = parameters;
    Py_XDECREF(nickName);
    PyObject *o = PyImport_AddModule("raw");
    PyObject *p = PyModule_GetDict(o);
    nickName = PyString_FromString((char *)nick.chars());
    PyDict_SetItemString(p, "nickname", nickName);
  }
}

void
Parser::parseKill(String from, String parameters)
{
  parameters = parameters.after("!");
  Util::do_on(3, "KILL", from.chars(), parameters.chars());
}

void
Parser::parseCommand(String window, String message) {
  String command;
  PyObject *o, *p;
  Window *w;
  char buf[8192];
  int n;

  Py_XDECREF(windowName);
  windowName = PyString_FromString((char *)window.chars());
  o = PyImport_AddModule("raw");
  p = PyModule_GetDict(o);
  PyDict_SetItemString(p, "window", windowName);

  if (message == "") return;
  if (message[0]=='/')
    message = message.after((int)0);
  if (message[0] == ' ') return;
  command = message.before(" ");
  if (command == "") command = message;
  command = downcase(command);
  message = message.after(" ");

//  if (command == "load") {
//    mainWindow->loadScript(message);
//    return;
//  }

  w = mainWindow->findWindow(window);
  o = PyDict_GetItem(globalDict,
        PyString_FromString((char *)command.chars()));

  if (o == 0) {
    w->insertLine(String("*** ") + command +
      " : No such function");
    return;
  }
  if (!PyCallable_Check(o)) {
    w->insertLine(String("*** ") + command +
      " is not a function");
    return;
  }

  strcpy(buf,command);
  strcat(buf, "(");
  n = Util::getArgNumber(command);
  for (int i = 0; i < n; i++) {

    if (i != 0) strcat(buf, ",");

    if (message == "") {
      strcat(buf, "''");
      continue;
    }

    if (message.before(" ") == "") {
      strcat(buf,"'");
      strcat(buf, Util::escapeString(message));
      strcat(buf,"'");
      message = "";
      continue;
    }

    strcat(buf,"\"");
    if (i == n-1)
      strcat(buf, Util::escapeString(message));
    else
      strcat(buf, Util::escapeString(message.before(" ")));
    strcat(buf,"\"");
    message = message.after(" ");
    message = Util::removeLeadingBlanks(message);
  }
  strcat(buf,")");
  PyRun_SimpleString(buf);
}

void
Parser::parse324(String from, String parameters)
{
  String channel, modes, params;

  params = parameters.after(" ");
  channel = downcase(params.before(" "));
  modes = params.after(" ");
  ChannelWindow *w = (ChannelWindow *)mainWindow->findWindow(channel);
  if (w != 0)
    w->parseMode(modes.before(" "), modes.after(" "));

  Util::do_on(2, "332", parameters.chars());
}

void
Parser::parse332(String from, String parameters)
{
  String channel;
  ChannelWindow *w;

  // parameters = "channel :topic"
  channel = parameters.after(" ");
  channel = downcase(channel.before(" "));
  w = (ChannelWindow *)mainWindow->findWindow(channel);
  w->setTopic(parameters.after(":"));

  Util::do_on(2, "332", parameters.chars());
}

void
Parser::parse353(String from, String parameters)
{
  ChannelWindow *w;
  String channel, param, nick;

  param = parameters.after(" ");
  param = param.after(" "); // We skip "= " or "* " or "@ "
  channel = downcase(param.before(" "));
  param = param.after(":");

  w = (ChannelWindow *)mainWindow->findWindow(channel);

  if (w!=0 && w->getStatus() == ChannelWindow::JOINING)
    while (1) {
      nick = param.before(" ");
      param = param.after(" ");
      if (nick == "") break;
      w->addNick(nick);
    }

  Util::do_on(2, "353", parameters.chars());  
}

void
Parser::parse364(String, String parameters)
{
  String from, to, params;
  params = parameters.after(" ");
  from = params.before(" ");
  to = params.after(" ");
  to = to.before(" ");

  if (mainWindow->linksWindow == 0)
    mainWindow->linksWindow = new LinksWindow();

  mainWindow->linksWindow->addLink(from, to);

  Util::do_on(2, "364", parameters.chars());  
}

void
Parser::parse365(String from, String parameters)
{
  mainWindow->linksWindow->doneLinks();
  Util::do_on(2, "365", parameters.chars());  
}

void
Parser::parse366(String from, String parameters)
{
  String channel;
  ChannelWindow *w;

  // parameters = "nick channel :End of /NAMES list."
  channel = parameters.after(" ");
  channel = downcase(channel.before(" "));
  w = (ChannelWindow *)mainWindow->findWindow(channel);
  if (w != 0)
    w->setStatus(ChannelWindow::AFTER_NAMES);

  Util::do_on(2, "366", parameters.chars());  
}

void
Parser::parse432(String from, String parameters)
{
  if (mainWindow->connected && !mainWindow->registered) {
    String n = Util::getString("Nickname change",
                 "Enter a valid nickname", FALSE, mainWindow);
    Util::send(socket,"NICK " + n);
  }

  Util::do_on(2, "432", parameters.chars());
}

void
Parser::parse433(String from, String parameters)
{
  if (mainWindow->connected && !mainWindow->registered) {
    String n = Util::getString("Nickname change",
                 "Enter a new nickname", FALSE, mainWindow);
    Util::send(socket,"NICK " + n);
  }

  Util::do_on(2, "433", parameters.chars());
}

