/*
 *  PyIRC.cpp - The main object.
 *  Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Python.h>

#include <GetOpt.h>
#include <fstream.h>
#include <Integer.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>

#include <qmsgbox.h>
#include <qpopmenu.h>
#include <qkeycode.h>
#include <qapp.h>
#include <qmlined.h>

#include "Window.h"
#include "FontSelector.h"
#include "ConnectDialog.h"
#include "PyIRC.h"
#include "Util.h"
#include "Defines.h"

String PyIRC::configFile = String(getpwuid(getuid())->pw_dir) +
                           "/" + CONFIG_FILE;
String PyIRC::bookFile   = String(getpwuid(getuid())->pw_dir) +
                           "/" + BOOK_FILE;
String PyIRC::initFile   = String(getpwuid(getuid())->pw_dir) +
                           "/" + INIT_FILE;

PyIRC::PyIRC(QWidget *parent, const char *name)
  : Window("-"), fileConnect(0), linksWindow(0), readSocketNotifier(0),
    writeSocketNotifier(0), parser(0), mode(0), idle(time(0))
{
  QPopupMenu *file = new QPopupMenu;
  CHECK_PTR(file);
  file->insertItem("&Connect...", this, SLOT(FileConnect()));
  file->insertItem("&Disconnect", this, SLOT(FileDisconnect()));
  file->insertSeparator();
  file->insertItem("&Quit", this, SLOT(FileQuit()), CTRL+Key_Q);

  QPopupMenu *options = new QPopupMenu;
  CHECK_PTR(options);
  options->insertItem("&Colors...", this, SLOT(OptionsColors()));
  options->insertItem("&Font...", this, SLOT(OptionsFont()));

  channelBookmarks = new QPopupMenu;
  channelBookmarks->insertItem("&Add bookmark...");
  channelBookmarks->insertSeparator();
  connect(channelBookmarks, SIGNAL(activated(int)), this,
          SLOT(channelBookmarkSelected(int)));

  queryBookmarks = new QPopupMenu;
  queryBookmarks->insertItem("&Add bookmark...");
  queryBookmarks->insertSeparator();
  connect(queryBookmarks, SIGNAL(activated(int)), this,
          SLOT(queryBookmarkSelected(int)));

  QPopupMenu *bookmarks = new QPopupMenu;
  CHECK_PTR(bookmarks);
  bookmarks->insertItem("&Channels", channelBookmarks);
  bookmarks->insertItem("&Queries", queryBookmarks);
  bookmarks->insertSeparator();
  bookmarks->insertItem("&Save bookmarks", this, SLOT(BookmarksSave()));

  QPopupMenu *dcc = new QPopupMenu;
  CHECK_PTR(dcc);
  dcc->insertItem("&Send...", this, SLOT(DccSend()));
  dcc->insertItem("&Chat...", this, SLOT(DccChat()));
  dcc->insertSeparator();
  dcc->insertItem("&Options...", this, SLOT(DccOptions()));

  QPopupMenu *help = new QPopupMenu;
  CHECK_PTR(help);
  help->insertItem("&About PyIRC...", this, SLOT(HelpAbout()));
  help->insertSeparator();
  help->insertItem("&License", this, SLOT(HelpLicense()));
  help->insertItem("&Using PyIRC", this, SLOT(HelpUsing()));
  help->insertItem("&Writing scripts", this, SLOT(HelpScripts()));

  menu = new QMenuBar(this);
  CHECK_PTR(menu);
  menu->setGeometry(0, 0, 500,30);
  menu->insertItem("&File", file);
  menu->insertItem("&Options", options);
  menu->insertItem("&Bookmarks", bookmarks);
  menu->insertItem("&DCC", dcc);
  menu->insertSeparator();
  menu->insertItem("&Help", help);

  newTitle();
  windowList.setAutoDelete(TRUE);
  connected = registered = FALSE;

  loadBookmarks();

  connect(messageEdit, SIGNAL(textChanged(const char*)), this,
      SLOT(update_idle()));
}

PyIRC::~PyIRC()
{
  if (registered)
    Util::send(readSocketNotifier->socket(),
               "QUIT :I use PyIRC");
  delete linksWindow;
  delete menu;
  delete channelBookmarks;
  delete queryBookmarks;
  delete fileConnect;
  delete readSocketNotifier;
  delete writeSocketNotifier;
}

void
PyIRC::parseCommandLine(int argc, char *argv[]) {
  int opt;
  GetOpt getopt(argc, argv, "hb:f:");

  while ((opt = getopt()) != EOF)
    switch (opt) {
      case 'f': configFile = getopt.optarg; break;
      case 'b': bookFile = getopt.optarg; break;
      case 'h': Util::giveHelp();
      case '?': Util::printUsage();
    }
}

void
PyIRC::loadBookmarks(void) {
  String s;
  QPopupMenu *target;
  ifstream file((const char *)bookFile);
  while (!file.eof()) {
    s = ""; readline(file, s);
    s = Util::removeLeadingBlanks(s);
    s = Util::removeTrailingBlanks(s);
    if (s == "") break;
    if (s[0] == '#' || s[0] == '&')
      target = channelBookmarks;
    else
      target = queryBookmarks;
    target->insertItem(s);
  }
}

void
PyIRC::channelBookmarkSelected(int index)
{
  if (!registered)
    return;

  if (index != 0) {
    Util::send(readSocketNotifier->socket(),
               "JOIN %s", channelBookmarks->text(index));
    return;
  }
  String channel = Util::getString("Channel bookmarks",
   "Enter the new bookmark:", FALSE, this);
  if (channel == "") return;
  if ((channel[0] != '#' && channel[0] != '&') || 
      (channel.index(' ') != -1)) {
    Util::showMessage("Error", "This is an invalid channel name", this);
    return;
  }
  channelBookmarks->insertItem(channel.chars());
}

void
PyIRC::queryBookmarkSelected(int index)
{
  Window *w;

  if (index != 0) {
    w = findWindow(queryBookmarks->text(index));
    if (w == 0) {
      w = (MessageWindow *)addMessageWindow(queryBookmarks->text(index));
      w->show();
    }
    return;
  }
  String nickname = Util::getString("Query bookmarks",
   "Enter the new bookmark:", FALSE, this);
  if (nickname == "") return;
  if (nickname.length() > 9 || nickname.index(' ') != -1) {
    Util::showMessage("Error", "This is an invalid nickname", this);
    return;
  }
  queryBookmarks->insertItem(nickname.chars());
}

void
PyIRC::BookmarksSave(void)
{
  ofstream file((const char *)bookFile, ios::out | ios::trunc);
  int i;

  for (i = 2; i < channelBookmarks->count(); i++)
    file << channelBookmarks->text(i) << endl;
  for (i = 2; i < queryBookmarks->count(); i++)
    file << queryBookmarks->text(i) << endl;
}

void
PyIRC::FileConnect() {
  if (fileConnect == 0) {
    fileConnect = new ConnectDialog(this);
    fileConnect->readConfigFile(configFile);
  }
  fileConnect->show();
}

void
PyIRC::FileDisconnect() {
  if (!connected) {
    insertLine("*** Not connected.");
    return;
  }
  ircDisconnect();
}

void
PyIRC::FileQuit() {
  if (QMessageBox::query("Confirmation",
                         "Are you sure you want to quit?", 0, 0, this))
    qApp->quit();
}

void
PyIRC::DccSend(void)
{
}

void
PyIRC::DccChat(void)
{
}

void
PyIRC::DccOptions(void)
{
}

void
PyIRC::HelpAbout() {
  Util::showMessage("About PyIRC",
                    "PyIRC\n\nCopyright (c) Etienne BERNARD 1997\n(eb"
                    " on IRCNet)\nThis software uses Qt, by "
                    "Troll Tech\n(http://www.troll.no)\nIt is distributed"
                    " under the GPL license.", this);
}

void
PyIRC::HelpLicense() {
  Util::showMessage("PyIRC License information",
"This software is

Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr).

It uses the GNU General Public License Version 2, that you
should have received with the source code of this program
(the file named \"LICENSE\"). If not, you can write to the:

Free Software Fundation, Inc.
675 Mass Ave
Cambridge, MA 02139, USA", this);
}

void
PyIRC::HelpUsing() {
}

void
PyIRC::HelpScripts() {
}

void
PyIRC::closedChild(Window *w)
{
  Util::do_on(2, "WINDOW_KILL", (w->getName()).chars());
  if (w->getName() != "-")
    windowList.remove(w);
  else
    FileQuit();
}

void
PyIRC::update_idle(void)
{
  idle = time(0);
}

time_t
PyIRC::idle_time(void)
{
  return idle;
}

void
PyIRC::closeEvent(QCloseEvent *e)
{
  if (QMessageBox::query("Confirmation",
      "Are you sure you want to quit?", 0, 0, this))
    e->accept();
  else
    e->ignore();
}

void
PyIRC::setNickname(String s)
{
  nick = s;
}

void
PyIRC::setUsername(String s)
{
  username = s;
}

void
PyIRC::setIrcname(String s)
{
  ircname = s;
}

void
PyIRC::setServer(Server *s)
{
  server.name = s->name;
  server.port = s->port;

  if (connected)
    ircDisconnect();
}

void
PyIRC::ircConnect(void)
{
  struct hostent *host;
  struct sockaddr_in addr;
  int sock; long arg = O_NONBLOCK;

  if (readSocketNotifier != 0) {
    ::close(readSocketNotifier->socket());
    delete parser;
  }

  connected = registered = FALSE;

  insertLine("*** Connecting to port " + String(dec(server.port)) +
              " of server " + server.name );

  if ((host=gethostbyname(server.name))!=0) {
    memset(&addr,0,sizeof(addr));
    addr.sin_family=host->h_addrtype;
    memcpy(&addr.sin_addr,host->h_addr,host->h_length);
    addr.sin_port=htons(server.port);
    if ((sock=socket(AF_INET,SOCK_STREAM,0))>=0)
      if (::connect(sock, (struct sockaddr *)&addr,sizeof(addr))>=0)
        if (fcntl(sock, F_SETFL, arg) >= 0) {
          connected = TRUE;
        }
  }  

  if (!connected) {
    insertLine("*** Not connected.");
    return;
  }

  readSocketNotifier = new QSocketNotifier(sock, QSocketNotifier::Read);
  writeSocketNotifier = new QSocketNotifier(sock, QSocketNotifier::Write);
  parser = new Parser(sock);

  QObject::connect(readSocketNotifier, SIGNAL(activated(int)),
                   this, SLOT(readData()));
  QObject::connect(writeSocketNotifier, SIGNAL(activated(int)),
                   this, SLOT(writeData()));
}

void
PyIRC::readData()
{
  String str;

  str = Util::receive(readSocketNotifier->socket());
#ifdef DBG
  cout << str << endl;
#endif

  if (str.length() == 0)
    ircDisconnect();
  else
    parser->parseInput(str);
}

void
PyIRC::writeData()
{
  Util::send(writeSocketNotifier->socket(),
             "NICK %s", (const char *)nick);
  Util::send(writeSocketNotifier->socket(),
             "USER %s . . :%s", (const char *)username,
             (const char *)ircname);
  writeSocketNotifier->setEnabled(FALSE);
  insertLine("*** Connection established.");
}

void
PyIRC::handleInput()
{
  String message = messageEdit->text();

  if (message == "")
    return;

  messageEdit->setText("");

  if (message[0]=='/')
    Parser::parseCommand("-", message);
}

Window *
PyIRC::findWindow(String n)
{
  unsigned int i;
  Window * w;
  QListIterator<Window> it(windowList);

  if (n == "-")
    return this;

  for (; w = it.current(); ++it) {
    if (w->getName() == n)
      return w;
  }
  return 0;
}

MessageWindow *
PyIRC::addMessageWindow(String n)
{
  MessageWindow *w = new MessageWindow(n);
  windowList.append(w);
  connect(w, SIGNAL(windowClosed(Window *)),
          this, SLOT(closedChild(Window *)));
  return w;
}

ChannelWindow *
PyIRC::addChannelWindow(String n)
{
  ChannelWindow *w = new ChannelWindow(n);
  windowList.append(w);
  connect(w, SIGNAL(windowClosed(Window *)),
          this, SLOT(closedChild(Window *)));
  return w;
}

bool
PyIRC::isOnChannel(String c)
{
  return (findWindow(c) != 0);
}

void
PyIRC::parseMode(String s) {
  char sign = '-';
  for (unsigned int i = 0; i < s.length(); i++)
    switch (s[i]) {
      case '-':
      case '+':
        sign = s[i]; break;
      case 'o':
        sign == '+' ? mode |= IRCOP : mode &= ~IRCOP; break;
      case 'w':
        sign == '+' ? mode |= WALLOPS : mode &= ~WALLOPS; break;
      case 's':
        sign == '+' ? mode |= SERVNOTICES : mode &= ~SERVNOTICES; break;
      case 'i':
        sign == '+' ? mode |= INVISIBLE : mode &= ~INVISIBLE; break;
    }
  newTitle();
}

void
PyIRC::newTitle(void)
{
  String t = "PyIRC - Control Window";
  if (mode != 0) {
    t += " - User mode +";
    if (mode & INVISIBLE) t += "i";
    if (mode & IRCOP) t += "o";
    if (mode & SERVNOTICES) t += "s";
    if (mode & WALLOPS) t += "w";
  }
  setCaption(t);
}

void
PyIRC::ircDisconnect(void)
{
  connected = registered = FALSE;
  ::close(readSocketNotifier->socket());
  ::close(writeSocketNotifier->socket());
  delete readSocketNotifier;
  delete writeSocketNotifier;
  delete parser;
  readSocketNotifier = 0;
  writeSocketNotifier = 0;
  parser = 0;

  insertLine("*** Connection closed.");
  Util::do_on(2, "DISCONNECT", "");
}

int
PyIRC::getSock(void)
{
  if (!connected)
    return -1;
  else
    return readSocketNotifier->socket();
}

void
PyIRC::loadScript(String filename)
{
  if (filename == "")
    return;

  if (access(filename.chars(), R_OK) != 0) {
    cout << "I can not access to " << filename << endl;
    return;
  }

  FILE *f = fopen(filename.chars(), "r");
  PyRun_SimpleFile(f, (char *)filename.chars());
  fclose(f);
}

void
PyIRC::loadInitFile(void)
{
  static int done = 0;

  if (done++) return;

  if (access(initFile, R_OK) == 0)
    loadScript(initFile);
}
