/*
 *  Util.cpp - Miscellaneous utility functions
 *  Copyright (c) 1997 Etienne Bernard (eb@via.ecp.fr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <Python.h>
#include <cgensupport.h>

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <iostream.h>
#include <qapp.h>
#include <qmsgbox.h>

#include "Util.h"
#include "AskDialog.h"
#include "Defines.h"

extern QApplication *app;
extern PyObject *globalDict;

void
Util::showMessage(const String title, const String text, QWidget *parent)
{
  QMessageBox mb(parent);
  mb.setText(text);
  mb.setButtonText("Ok");
  mb.setCaption(title);
  mb.show();
}

String
Util::escapeString(const String s)
{
  String r = "";
  unsigned int i;

  for (i=0; i < s.length(); i++) {
    switch (s[i]) {
      case '\'':
      case '\"':
      case '\\':
        r = r + "\\" + s[i]; break;
      default:
        r += s[i];
    }
  }
  return r;
}

String
Util::convertToAlpha(const String s)
{
  String r="";
  unsigned int i;
  int pos = 0, beeped = 0;

  for (i=0; i < s.length(); i++) {
    switch (s[i]) {
      case '\002':  /* CTRL-B : Bold */
        r += "@b";
        break;
      case '\007':  /* CTRL-G : Beep - Only one beep per line */
        if (!beeped) {
          app->beep();
          beeped = 1;
        }
        break;
      case '\026':  /* CTRL-V : Reverse */
        r += "@r";
        break;
      case '\037':  /* CTRL-_ : Underline */
        r += "@u";
        break;
      case '@':
        r += "@@";
        break;
      case '\t':
       r = r + String("        ").before(8 - pos % 8);
        pos = pos + 8 - pos % 8;
        break;
      default:
        r += s[i]; pos++;
    }
  }
  return r;
}

String
Util::removeNonAlpha(String s)
{
  String r;
  unsigned int i;

  for (i = 0; i < s.length(); i++)
    if ((s[i]>=32 && s[i]<=126) || ((unsigned char)s[i] >= 160))
//    if (((char)s[i] >= 32 && (char)s[i] <= 126) || ((char)s[i] >= 160))
      r = r + s[i];

  return r;
}

void
Util::giveHelp(void)
{
  cout << "PyIRC is a graphical IRC client, which uses (WILL use) the Python programming" << endl;
  cout << "language for scripting. This software uses the Qt library, made by" << endl;
  cout << "Troll Tech (see http://www.troll.no/ for information about Qt)." << endl << endl;

  cout << "Usage:" << endl;
  cout << "   PyIRC [-h] [-f configfile]" << endl;

  exit(0);
}

void
Util::printUsage(void)
{
  cout << "Usage:" << endl;
  cout << "   PyIRC [-h] [-f configfile]" << endl;

  exit(1);
}

String
Util::removeLeadingBlanks(String s) {
  unsigned i = 0;

  while ((i < s.length()) && ((s[i]==' ') || (s[i]=='\t')))
    i++;

  return s.after(int(i-1));
}

String
Util::removeTrailingBlanks(String s) {
  int i = s.length() - 1;

  while ((i > 0) && ((s[i]==' ') || (s[i]=='\t')))
    i--;

  return s.before(i+1);
}

void
Util::send(int socket, const char *format, ...)
{
  va_list msg;
  char buf[8192];

  va_start(msg, format);
  vsprintf(buf, format, msg);
  va_end(msg);
#ifdef DBG
  cout << "======> " << buf << endl;
#endif
  write(socket, buf, strlen(buf));
  write(socket, "\r\n", 2);
}

String
Util::receive(int socket)
{
  char buf[1024];
  int pos = 0;
  char r;

  do {
    if (read(socket, &r, 1) <= 0)
      return String("");

    buf[pos++]=r;

  } while (r != '\n');

  buf[pos-2] = '\0';

  return String(buf);
}

bool
Util::isServer(String s)
{
  return (s.index("@") == -1);
}

String
Util::timeStr(time_t t)
{
  time_t t2 = t;
  String ts = ctime(&t2);
  ts = ts.before("\n");
  return ts;
}

String
Util::getString(String title, String text, bool pass, QWidget *parent)
{
  AskDialog d(title, text, pass, parent);
  if (d.exec())
    return String(d.getResult());
  else
    return String("");
}

int
Util::getArgNumber(String command)
{
  PyObject *o, *p;
  long argNumber;

  o = PyDict_GetItem(globalDict, PyString_FromString((char *)command.chars()));
  if (o == 0)
    return -1;
  if (!PyObject_HasAttrString(o, "func_code"))
    return -1;

  p = PyObject_GetAttr(o, PyString_FromString("func_code"));
  o = PyObject_GetAttr(p, PyString_FromString("co_argcount"));
  Py_DECREF(p);
  PyArg_GetLong(o, 1, 0, &argNumber);
  Py_DECREF(o);
  return argNumber;
}

int
Util::do_on(int num, ...)
{
  va_list args;
  va_start(args, num);
  char s[256] = "(";
  long res;

  for (int i = 0; i < num; i++)
    strcat(s, "s");
  strcat(s,")");

  PyObject *arglist, *result, *function;
  function = PyImport_AddModule("pyirc");
  function = PyModule_GetDict(function);
  function = PyDict_GetItemString(function, "do_on");

  arglist = Py_VaBuildValue(s, args);

  if (function == 0 || arglist == 0)
    return 0;

  result = PyEval_CallObject(function, arglist);
  Py_DECREF(arglist);

  if (result == 0)
    return 0;

  PyArg_GetLong(result, 1, 0, &res);
  Py_DECREF(result);

  return res;
}

bool
Util::isChannelName(String s)
{
  if (s[0] != '#' && s[0] != '&')
    return FALSE;
  return TRUE;
}
